@@ -8,8 +8,9 @@ test_that("st_make_valid works", {
 
    	y = st_make_valid(x)
 	expect_true(st_is_valid(y))
-   	expect_true(st_is_valid(lwgeom::st_make_valid(x[[1]])))
-   	expect_true(st_is_valid(lwgeom::st_make_valid(st_sf(a = 1, geom = x))))
+   	expect_true(st_is_valid(st_make_valid(x[[1]])))
+   	expect_true(st_is_valid(lwgeom_make_valid(x)))
+   	expect_true(st_is_valid(st_make_valid(st_sf(a = 1, geom = x))))
 	expect_equal(lwgeom::st_geohash(st_sfc(st_point(c(1.5,3.5)), st_point(c(0,90))), 2), c( "s0","up"))
 	expect_equal(lwgeom::st_geohash(st_sfc(st_point(c(1.5,3.5)), st_point(c(0,90))), 10),
 		c("s095fjhkbx","upbpbpbpbp"))
@@ -22,6 +23,10 @@ test_that("st_make_valid works", {
 	p1 = st_point(c(7,52))
 	geom.sf = st_sfc(p1, crs = 4326)
 	x <- st_transform_proj(geom.sf, "+proj=wintri")
+	if (sf_extSoftVersion()["proj.4"] >= "6.0.0") {
+		x2 <- st_transform_proj(geom.sf, c("EPSG:4326", "+proj=wintri"))
+	}
+	x3 <- st_transform_proj(geom.sf, st_crs(3857))
 	p = st_crs(4326)$proj4string
 	x <- st_transform_proj(structure(geom.sf[[1]], proj4string = p), "+proj=wintri")
 	nc = st_read(system.file("gpkg/nc.gpkg", package="sf"), quiet = TRUE)
